// zoomasm -- zoom video assembler
// (c) 2019,2020,2023 Claude Heiland-Allen
// SPDX-License-Identifier: AGPL-3.0-or-later

#include "utility.h"

#include <algorithm>
#include <cstring>
#include <cstdlib>

#if __cplusplus >= 202002L
#define CPlusPlus20
#endif

extern bool starts_with(const std::string &haystack, const std::string &needle)
{
#ifdef CPlusPlus20
  return haystack.starts_with(needle);
#else
  // https://stackoverflow.com/a/40441240
  return haystack.rfind(needle, 0) == 0;
#endif
}

extern bool ends_with(const std::string &haystack, const std::string &needle)
{
#ifdef CPlusPlus20
  return haystack.ends_with(needle);
#else
  // https://stackoverflow.com/a/42844629
  return haystack.size() >= needle.size() && 0 == haystack.compare(haystack.size() - needle.size(), needle.size(), needle);
#endif
}

extern bool contains(const std::set<std::string> &haystack, const std::string &needle)
{
#ifdef CPlusPlus20
  return haystack.contains(needle);
#else
  // https://stackoverflow.com/a/1701083
  return haystack.find(needle) != haystack.end();
#endif
}

extern std::set<std::string> intersection(const std::set<std::string> &s1, const std::set<std::string> &s2)
{
  // https://stackoverflow.com/a/13448094
  std::set<std::string> s3;
  std::set_intersection(s1.begin(), s1.end(), s2.begin(), s2.end(), std::inserter(s3,s3.begin()));
  return s3;
}

extern char *mystrdup(const char* s)
{
  size_t slen = std::strlen(s);
  char* result = (char *) std::malloc(slen + 1);
  if (result == NULL)
  {
    return NULL;
  }
  std::memcpy(result, s, slen + 1);
  return result;
}

extern std::istream& getline(std::istream& is, std::string& t)
{
  // https://stackoverflow.com/a/6089413

    t.clear();

    // The characters in the stream are read one-by-one using a std::streambuf.
    // That is faster than reading them one-by-one using the std::istream.
    // Code that uses streambuf this way must be guarded by a sentry object.
    // The sentry object performs various tasks,
    // such as thread synchronization and updating the stream state.

    std::istream::sentry se(is, true);
    std::streambuf* sb = is.rdbuf();

    for(;;) {
        int c = sb->sbumpc();
        switch (c) {
        case '\n':
            return is;
        case '\r':
            if(sb->sgetc() == '\n')
                sb->sbumpc();
            return is;
        case std::streambuf::traits_type::eof():
            // Also handle the case when the last line has no line ending
            if(t.empty())
                is.setstate(std::ios::eofbit);
            return is;
        default:
            t += (char)c;
        }
    }
}
